package cz.fidentis.analyst.symmetry;

import cz.fidentis.analyst.gui.canvas.Canvas;
import cz.fidentis.analyst.gui.scene.DrawablePlane;
import cz.fidentis.analyst.symmetry.SignificantPoints.CurvatureAlg;
import java.awt.event.ActionEvent;
import javax.swing.AbstractAction;
import javax.swing.JPanel;
import javax.swing.JToggleButton;

/**
 * Action listener for the manipulation with the symmetry plane.
 * 
 * @author Radek Oslejsek
 */
public class SymmetryPlaneAction extends AbstractAction {
    
    public static final String ACTION_COMMAND_ON_OFF = "on-off";
    
    private final SymmetryPanel controlPanel = new SymmetryPanel();
    private final JPanel topControlPanel;
    private final Canvas canvas;
    
    private final SymmetryConfig config = new SymmetryConfig();
    private CurvatureAlg curvatureAlg = CurvatureAlg.GAUSSIAN;
    
    /**
     * Constructor.
     * 
     * @param topControlPanel Top component for placing control panels
     * @param canvas OpenGL canvas
     */
    public SymmetryPlaneAction(JPanel topControlPanel, Canvas canvas) {
        this.canvas = canvas;
        this.topControlPanel = topControlPanel;
        this.topControlPanel.add(controlPanel);
        controlPanel.setVisible(false);
    }
    
    @Override
    public void actionPerformed(ActionEvent ae) {
        String action = ae.getActionCommand();
        switch (action) {
            case ACTION_COMMAND_ON_OFF:
                JToggleButton button = (JToggleButton) ae.getSource();
                DrawablePlane plane = canvas.getScene().getDrawableSymmetryPlane(0);
                if (button.isSelected()) {
                    controlPanel.setVisible(true);
                    if (plane == null) {
                        recomputeSymmetryPlane();
                    }
                    plane.show();
                } else {
                    controlPanel.setVisible(false);
                    if (plane != null) {
                        plane.hide();
                    }
                }
                break;
            default:
                throw new UnsupportedOperationException(action);
        }
        canvas.renderScene();
    }
    
    private void recomputeSymmetryPlane() {
        //setCursor(Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR));
        SymmetryEstimator se = new SymmetryEstimator(this.config, this.curvatureAlg);
        canvas.getScene().getDrawableFace(0).getModel().compute(se);
        DrawablePlane plane = new DrawablePlane(se.getSymmetryPlaneMesh(), se.getSymmetryPlane());
        canvas.getScene().setDrawableSymmetryPlane(0, plane);
    }
        
}
